/********************************************************************************
*                                                                               *
*                      F X   M O V I E   M A N A G E R                          *
*                                                                               *
*********************************************************************************
* Copyright(C) 2008-2012 by Luigi Ziviani <ziviani_luigi@yahoo.it>.             *
* All Rights Reserved.                                 v 6.1                    *
*********************************************************************************
* Id: fxplayer_buffer.h,                               2008/01/01               *
*********************************************************************************
* This file is part of FXMOVIEMANAGER                                           *
*                                                                               *
* FXMOVIEMANAGER is free software; you can redistribute it and/or modify        *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation; either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* FXMOVIEMANAGER is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with this program; if not, write to the Free Software                   *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     * 
********************************************************************************/

               /********************************************
               * Acknowledgements:                         *
               * the code written in this file is based on * 
               * Various utilities for command line tools  *
               * Copyright (c) 2000-2003 Fabrice Bellard,  *
               * that are part of FFmpeg.                  *
               ********************************************/

#ifndef FXPLAYER_BUFFER_H
#define FXPLAYER_BUFFER_H

#include <stdint.h>

#include <libavcodec/avcodec.h>
#include <libavfilter/avfilter.h>
#include <libavformat/avformat.h>
#include <libswscale/swscale.h>

#ifdef __MINGW32__
#undef main /* We don't want SDL to override our main() */
#endif

typedef struct FrameBuffer {
    uint8_t *base[4];
    uint8_t *data[4];
    int  linesize[4];
    int h, w;
    enum PixelFormat pix_fmt;
    int refcount;
    struct FrameBuffer **pool;      ///< head of the buffer pool
    struct FrameBuffer *next;
} FrameBuffer;



/**
 * Get a frame from the pool. This is intended to be used as a callback for
 * AVCodecContext.get_buffer.
 *
 * @param s codec context. s->opaque must be a pointer to the head of the
 *          buffer pool.
 * @param frame frame->opaque will be set to point to the FrameBuffer
 *              containing the frame data.
 */
int codec_get_buffer(AVCodecContext *s, AVFrame *frame);

/**
 * A callback to be used for AVCodecContext.release_buffer along with
 * codec_get_buffer().
 */
void codec_release_buffer(AVCodecContext *s, AVFrame *frame);

/**
 * A callback to be used for AVFilterBuffer.free.
 * @param fb buffer to free. fb->priv must be a pointer to the FrameBuffer
 *           containing the buffer data.
 */
void filter_release_buffer(AVFilterBuffer *fb);

/**
 * Free all the buffers in the pool. This must be called after all the
 * buffers have been released.
 */
void free_buffer_pool(FrameBuffer **pool);

#endif /* FXPLAYER_BUFFER_H */
